/*
 *  Text screen control library
 *  
 *  Ver. 1.0  1996.04.28  first release.
 *  Ver. 1.1  1997.09.28  add tgetstr() NULL check
 *  Ver. 2.0  2000.02.24  GPL version
 */
#include <stdio.h>
#include <stdlib.h>
#if defined(MINIX)
#include <sgtty.h>
#else
#include <termio.h>
#endif
#include <termcap.h>
#include <string.h>
#include <ctype.h>
#if !defined(MINIX)
#include <sys/file.h>
#endif

#include "tscrn.h"

#define FD_STDIN   0
#define FD_STDOUT  1
#define FD_STDERR  2

#define	FALSE   0
#define TRUE    !FALSE

#define DEL ((unsigned char)0x7f)

static char NULLSTR[1] = {'\0'};

static char Termbuf[1024];
static char *CL, *CM;
static char cap[1024];
static int LINES, COLS;

#if 0
static int outchar();
#endif

static unsigned char *T_BOLD_p;
static unsigned char *T_REVERSE_p;
static unsigned char *T_UNDERLINE_p;
static unsigned char *T_NORMAL_p;
static unsigned char *T_BLACK_p;
static unsigned char *T_RED_p;
static unsigned char *T_GREEN_p;
static unsigned char *T_YELLOW_p;
static unsigned char *T_BLUE_p;
static unsigned char *T_PURPLE_p;
static unsigned char *T_CYAN_p;
static unsigned char *T_WHITE_p;

static unsigned char *T_FC_p;
static unsigned char *T_FC2_p;
static unsigned char *T_BC_p;
static unsigned char *T_BC2_p;
static unsigned char *T_DC_p;
static unsigned char *T_DC2_p;
static unsigned char *T_IC_p;
static unsigned char *T_IC2_p;
static unsigned char *T_BS_p;
static unsigned char *T_BS2_p;
static unsigned char *T_CHAR_p;
static unsigned char *T_WCHAR_p;
static unsigned char *T_IGNORE_p;
static unsigned char *T_PI_p;
static unsigned char *T_PI2_p;
static unsigned char *T_NI_p;
static unsigned char *T_NI2_p;
static unsigned char *T_PU_p;
static unsigned char *T_PU2_p;
static unsigned char *T_PD_p;
static unsigned char *T_PD2_p;
static unsigned char *T_ESC_p;
static unsigned char *T_TAB_p;
static unsigned char *T_BTAB_p;
static unsigned char *T_ENT_p;
static unsigned char *T_F01_p;
static unsigned char *T_F02_p;
static unsigned char *T_F03_p;
static unsigned char *T_F04_p;
static unsigned char *T_F05_p;
static unsigned char *T_F06_p;
static unsigned char *T_F07_p;
static unsigned char *T_F08_p;
static unsigned char *T_F09_p;
static unsigned char *T_F10_p;
static unsigned char *T_F11_p;
static unsigned char *T_F12_p;

#define EUC_SS2 0x8e
#define EUC_SS3 0x8f

#define ASCii 0
#define JIS   1
#define EUC   2
#define SJIS  3
static int CharSet;

static char  scrnbuf[T_MAXSCRNBUF];

extern int  ioctl();
extern int  write();
extern int  read();
extern char *fgets();

static int TGetChar();
static int IsEnd();
static int IsWCharBuf();
static int ScratchTail();
static int TSetCursorItem();
static int StrCpy();
static int StrnCpy();
static int IsWChar();
static int IsChar();
static TITEM *TGetTitemPointer();
static TITEM *TGetForwTitemPointer();
static TITEM *TGetNextTitemPointer();
static TITEM *TGetBackTitemPointer();
static TITEM *TGetPrevTitemPointer();
static int TOutputClearScreen();
static int TOutputItem();

/*
 * int TInitTextLib()
 *
 * function : initialize libtext
 *
 * return :  0 normal
 *          -1 abnormal
 */
int TInitTextLib()
{
    char *term, *p, *cp;
    
#ifdef TIOCGWINSZ
    struct winsize winsize;
#endif
    
    if ((term = getenv("TERM")) == (char *)NULL){
	return -1;
    }
    if (tgetent(Termbuf, term) != 1){
	return -1;
    }
    
    LINES = tgetnum("li");
    COLS = tgetnum("co");
    
#ifdef TIOCGWINSZ
    if (ioctl(FD_STDOUT, TIOCGWINSZ, &winsize) != -1){
	LINES = winsize.ws_row;
	COLS = winsize.ws_col;
    }
#endif
    
    if ((p = getenv("LINES")) != (char *)NULL){
	LINES = atoi(p);
    }
    if ((p = getenv("COLS")) != (char *)NULL){
	COLS = atoi(p);
    }
    
    cp = cap;
#if !defined(MINIX)
    if ((p = tgetstr("pc", &cp)) != (char *)NULL){
	PC = *p;
    }
    BC = tgetstr("bc", &cp);
    if (BC == NULL){
	BC = NULLSTR;
    }
    UP = tgetstr("up", &cp);
    if (UP == NULL){
	UP = NULLSTR;
    }
#endif
    
    CL = tgetstr("cl", &cp);
    if (CL == NULL){
	CL = NULLSTR;
    }
    CM = tgetstr("cm", &cp);
    if (CM == NULL){
	CM = NULLSTR;
    }
    
    T_FC_p = tgetstr("kr", &cp);
    if (T_FC_p == NULL){
	T_FC_p = NULLSTR;
    }
    T_FC2_p = T_FC_STR;
    T_BC_p = tgetstr("kl", &cp);
    if (T_BC_p == NULL){
	T_BC_p = NULLSTR;
    }
    T_BC2_p = T_BC_STR;
    T_DC_p = tgetstr("kD", &cp);
    if (T_DC_p == NULL){
	T_DC_p = NULLSTR;
    }
    T_DC2_p = T_DC_STR;
    T_IC_p = tgetstr("kI", &cp);
    if (T_IC_p == NULL){
	T_IC_p = NULLSTR;
    }
    T_IC2_p = T_IC_STR;
    T_BS_p = tgetstr("kb", &cp);
    if (T_BS_p == NULL){
	T_BS_p = NULLSTR;
    }
    T_BS2_p = T_BS_STR;
    T_CHAR_p = (char *)NULL;
    T_WCHAR_p = (char *)NULL;
    T_IGNORE_p = (char *)NULL;
    T_PI_p = tgetstr("ku", &cp);
    if (T_PI_p == NULL){
	T_PI_p = NULLSTR;
    }
    T_PI2_p = T_PI_STR;
    T_NI_p = tgetstr("kd", &cp);
    if (T_NI_p == NULL){
	T_NI_p = NULLSTR;
    }
    T_NI2_p = T_NI_STR;
    T_PU_p = tgetstr("kP", &cp);
    if (T_PU_p == NULL){
	T_PU_p = NULLSTR;
    }
    T_PU2_p = T_PU_STR;
    T_PD_p = tgetstr("kN", &cp);
    if (T_PD_p == NULL){
	T_PD_p = NULLSTR;
    }
    T_PD2_p = T_PD_STR;
    T_ESC_p = T_ESC_STR;
    T_TAB_p = T_TAB_STR;
    T_BTAB_p = T_BTAB_STR;
    T_ENT_p = T_ENT_STR;
    T_F01_p = tgetstr("k1", &cp);
    if (T_F01_p == NULL){
	T_F01_p = NULLSTR;
    }
    T_F02_p = tgetstr("k2", &cp);
    if (T_F02_p == NULL){
	T_F02_p = NULLSTR;
    }
    T_F03_p = tgetstr("k3", &cp);
    if (T_F03_p == NULL){
	T_F03_p = NULLSTR;
    }
    T_F04_p = tgetstr("k4", &cp);
    if (T_F04_p == NULL){
	T_F04_p = NULLSTR;
    }
    T_F05_p = tgetstr("k5", &cp);
    if (T_F05_p == NULL){
	T_F05_p = NULLSTR;
    }
    T_F06_p = tgetstr("k6", &cp);
    if (T_F06_p == NULL){
	T_F06_p = NULLSTR;
    }
    T_F07_p = tgetstr("k7", &cp);
    if (T_F07_p == NULL){
	T_F07_p = NULLSTR;
    }
    T_F08_p = tgetstr("k8", &cp);
    if (T_F08_p == NULL){
	T_F08_p = NULLSTR;
    }
    T_F09_p = tgetstr("k9", &cp);
    if (T_F09_p == NULL){
	T_F09_p = NULLSTR;
    }
#if defined(MINIX)
    T_F10_p = tgetstr("kA", &cp);
#else
    T_F10_p = tgetstr("k0", &cp);
#endif
    if (T_F10_p == NULL){
	T_F10_p = NULLSTR;
    }
    T_F11_p = tgetstr("F1", &cp);
    if (T_F11_p == NULL){
	T_F11_p = NULLSTR;
    }
    T_F12_p = tgetstr("F2", &cp);
    if (T_F12_p == NULL){
	T_F12_p = NULLSTR;
    }

    T_BOLD_p = tgetstr("md", &cp);
    if (T_BOLD_p == NULL){
	T_BOLD_p = NULLSTR;
    }
    T_REVERSE_p = tgetstr("mr", &cp);
    if (T_REVERSE_p == NULL){
	T_REVERSE_p = NULLSTR;
    }
    T_UNDERLINE_p = tgetstr("us", &cp);
    if (T_UNDERLINE_p == NULL){
	T_UNDERLINE_p = NULLSTR;
    }
    T_NORMAL_p = tgetstr("me", &cp);
    if (T_NORMAL_p == NULL){
	T_NORMAL_p = NULLSTR;
    }
    T_BLACK_p = T_BLACK_STR;
    T_RED_p = T_RED_STR;
    T_GREEN_p = T_GREEN_STR;
    T_YELLOW_p = T_YELLOW_STR;
    T_BLUE_p = T_BLUE_STR;
    T_PURPLE_p = T_PURPLE_STR;
    T_CYAN_p = T_CYAN_STR;
    T_WHITE_p = T_WHITE_STR;

    if ((p = getenv("LANG")) != (char *)NULL || (p = getenv("LC_CTYPE")) != (char *)NULL){
	if (strcmp(p, "C") != 0){
	    CharSet = ASCii;
	}
	if (strcmp(p, "ja_JP.JIS") == 0){
	    CharSet = JIS;
	}
	if (strcmp(p, "ja_JP.jis7") == 0){
	    CharSet = JIS;
	}
	if (strcmp(p, "ja_JP.EUC") == 0){
	    CharSet = EUC;
	}
	if (strcmp(p, "ja_JP.ujis") == 0){
	    CharSet = EUC;
	}
	if (strcmp(p, "japan") == 0){
	    CharSet = EUC;
	}
	if (strcmp(p, "japanese") == 0){
	    CharSet = EUC;
	}
	if (strcmp(p, "ja_JP.SJIS") == 0){
	    CharSet = SJIS;
	}
	if (strcmp(p, "ja_JP.mscode") == 0){
	    CharSet = SJIS;
	}
    }else{
	CharSet = EUC;
    }

    return 0;
}

/*
 * int TIOString(col, row, len, str)
 *
 * function : input/output string
 *
 * argument : col display column(0 origin)
 *            row display row(0 origin)
 *            len input field length
 *            str string buffer
 *
 * return : action to terminate input
 *
 *        T_FC     right cursor
 *        T_BC     left cursor
 *        T_CHAR   spill over single byte character
 *        T_WCHAR  spill over double byte character
 *        T_PI     up cursor
 *        T_NI     down cursor
 *        T_PU     PageUp
 *        T_PD     PageDown
 *        T_ESC    ESCAPE
 *        T_TAB    TAB(^I)
 *        T_BTAB   ^U
 *        T_ENT    ENTER
 *        T_F01    FUNCTION01
 *        T_F02
 *        T_F03
 *        T_F04
 *        T_F05
 *        T_F06
 *        T_F07
 *        T_F08
 *        T_F09
 *        T_F10
 *        T_F11
 *        T_F12    FUNCTION12
 *        
 *        -1 abnormal
 */
int TIOString(col, row, len, str)
    int col;
    int row;
    int len;
    unsigned char *str;
{
    if (str == (unsigned char *)NULL){
	return -1;
    }
    TOutputString(col, row, len, str);
    return TInputString(col, row, len, str);
}
      
/*
 * int TOutputString(col, row, len, str)
 *
 * function : output string
 *
 * argument : col display column(0 origin)
 *            row display row(0 origin)
 *            len input field length
 *            str string buffer
 *
 * return :  0 normal
 *          -1 abnormal
 */
int TOutputString(col, row, len, str)
    int col;
    int row;
    int len;
    unsigned char *str;
{
    int i;
    
    if (str == (unsigned char *)NULL){
	return -1;
    }
    TSetCursor(col, row);
    write(FD_STDOUT, str, strlen(str));
    for (i = strlen(str); i < len; i++){
	write(FD_STDOUT, " ", 1);
    }
    return 0;
}

/*
 * int TInputString(col, row, len, str)
 *
 * function : input string
 *
 * argument : col display column(0 origin)
 *            row display row(0 origin)
 *            len input field length
 *            str string buffer
 *
 * return : action to terminate input
 *
 *        T_FC     right cursor
 *        T_BC     left cursor
 *        T_CHAR   spill over single byte character
 *        T_WCHAR  spill over double byte character
 *        T_PI     up cursor
 *        T_NI     down cursor
 *        T_PU     PageUp
 *        T_PD     PageDown
 *        T_ESC    ESCAPE
 *        T_TAB    TAB(^I)
 *        T_BTAB   ^U
 *        T_ENT    ENTER
 *        T_F01    FUNCTION01
 *        T_F02
 *        T_F03
 *        T_F04
 *        T_F05
 *        T_F06
 *        T_F07
 *        T_F08
 *        T_F09
 *        T_F10
 *        T_F11
 *        T_F12    FUNCTION12
 *        
 *        -1 abnormal
 */
int TInputString(col, row, len, str)
    int col;
    int row;
    int len;
    unsigned char *str;
{
#if defined(MINIX)
    struct sgttyb tio, tin;
#else
    struct termio tio, tin;
#endif
    unsigned char ch[3];
    int status, i, j, cp;
    
    if (str == (unsigned char *)NULL){
	return -1;
    }
#if defined(MINIX)
    ioctl(FD_STDIN, TIOCGETP, &tio);
    tin = tio;
    tin.sg_flags |= CBREAK;
    tin.sg_flags |= CRMOD;
    tin.sg_flags &= ~ECHO;
    ioctl(FD_STDIN, TIOCSETP, &tin);
#else
    ioctl(FD_STDIN, TCGETA, &tio);
    tin = tio;
    tin.c_lflag &= ~ECHO;
    tin.c_lflag &= ~ICANON;
    tin.c_cc[VMIN] = 1;
    tin.c_cc[VTIME] = 0;
    ioctl(FD_STDIN, TCSETA, &tin);
#endif
    
    TSetCursor(col, row);
    cp = 0;
    
    status = TGetChar(ch);
    while (!IsEnd(status)){
	switch (status){
	  case T_FC:
	    if (len == 0){
		return T_FC;
	    }
	    if (IsWCharBuf(str, cp)){
		cp++;
	    }
	    cp++;
	    if (len <= cp){
		ScratchTail(str);
		return T_FC;
	    }
	    TSetCursorItem(col, row, cp);
	    break;
	  case T_BC:
	    if (len == 0){
		return T_BC;
	    }
	    if (IsWCharBuf(str, cp - 2)){
		cp--;
	    }
	    cp--;
	    if (cp < 0){
		ScratchTail(str);
		return T_BC;
	    }
	    TSetCursorItem(col, row, cp);
	    break;
	  case T_DC:
	    if (len == 0){
		break;
	    }
	    if (IsWCharBuf(str, cp)){
		if (cp < strlen(str)){
		    StrCpy(str + cp, str + cp + 2, len);
		    write(FD_STDOUT, str + cp, strlen(str + cp));
		    write(FD_STDOUT, "  ", 2);
		    TSetCursorItem(col, row, cp);
		}
	    }else{
		if (cp < strlen(str)){
		    StrCpy(str + cp, str + cp + 1, len);
		    write(FD_STDOUT, str + cp, strlen(str + cp));
		    write(FD_STDOUT, " ", 1);
		    TSetCursorItem(col, row, cp);
		}
	    }
	    break;
	  case T_IC:
	    if (len == 0){
		break;
	    }
	    if (cp < strlen(str)){
		if (strlen(str) < len){
		    StrnCpy(str + cp + 1, str + cp, strlen(str) - cp, len);
		}else{
		    if (IsWCharBuf(str, len - 2)){
			StrnCpy(str + cp + 1, str + cp, strlen(str) - cp - 2, len);
			*(str + len - 1) = '\0';
		    }else{
			StrnCpy(str + cp + 1, str + cp, strlen(str) - cp - 1, len);
		    }
		}
		*(str + cp) = ' ';
		write(FD_STDOUT, str + cp, strlen(str + cp));
		TSetCursorItem(col, row, cp);
	    }
	    break;
	  case T_BS:
	    if (len == 0){
		break;
	    }
	    if (IsWCharBuf(str, cp - 2)){
		if (0 < cp){
		    cp -= 2;
		    TSetCursorItem(col, row, cp);
		    if (cp < strlen(str)){
			StrCpy(str + cp, str + cp + 2, len);
			write(FD_STDOUT, str + cp, strlen(str + cp));	
			write(FD_STDOUT, "  ", 2);
			TSetCursorItem(col, row, cp);
		    }
		}
	    }else{
		if (0 < cp){
		    cp--;
		    TSetCursorItem(col, row, cp);
		    if (cp < strlen(str)){
			StrCpy(str + cp, str + cp + 1, len);
			write(FD_STDOUT, str + cp, strlen(str + cp));	
			write(FD_STDOUT, " ", 1);
			TSetCursorItem(col, row, cp);
		    }
		}
	    }
	    break;
	  case T_CHAR:
	    if (len == 0){
		return T_CHAR;
	    }
	    if (strlen(str) <= cp && *ch == ' '){
		;
	    }else{
		if (strlen(str) <= cp){
		    *(str + cp + 1) = '\0';
		    for (i = strlen(str); i < cp; i++){
			*(str + i) = ' ';
		    }
		}
		if (IsWCharBuf(str, cp)){
		    *(str + cp) = *ch;
		    *(str + cp + 1) = ' ';
		    j = 2;
		}else{
		    *(str + cp) = *ch;
		    j = 1;
		}
		write(FD_STDOUT, str + cp, j);
	    }
	    cp++;
	    TSetCursorItem(col, row, cp);
	    if (len <= cp){
		ScratchTail(str);
		return T_CHAR;
	    }
	    break;
	  case T_WCHAR:
	    if (len == 0){
		return T_WCHAR;
	    }
	    if (cp < (len - 1)){
		if (strlen(str) <= cp){
		    *(str + cp + 2) = '\0';
		    for (i = strlen(str); i < cp; i++){
			*(str + i) = ' ';
		    }
		}
		if (IsWCharBuf(str, cp + 1)){
		    *(str + cp) = *ch;
		    *(str + cp + 1) = *(ch + 1);
		    *(str + cp + 2) = ' ';
		    j = 3;
		}else{
		    *(str + cp) = *ch;
		    *(str + cp + 1) = *(ch + 1);
		    j = 2;
		}
		write(FD_STDOUT, str + cp, j);
		cp += 2;
		TSetCursorItem(col, row, cp);
		if (len <= cp){
		    ScratchTail(str);
		    return T_WCHAR;
		}
	    }else{
		if (strlen(str) <= cp){
		    *(str + cp + 1) = '\0';
		    for (i = strlen(str); i < cp; i++){
			*(str + i) = ' ';
		    }
		}
		*(str + cp) = ' ';
		write(FD_STDOUT, str + cp, strlen(str + cp));
		cp++;
		TSetCursorItem(col, row, cp);
		ScratchTail(str);
		return T_WCHAR;
	    }
	    break;
	  case T_IGNORE:
	  default:
	    break;
	}
	status = TGetChar(ch);
    }
    
#if defined(MINIX)
    ioctl(FD_STDIN, TIOCSETP, &tio);
#else
    ioctl(FD_STDIN, TCSETA, &tio);
#endif
    
    ScratchTail(str);
    return status;
}

static int StrCpy(dst, src, len)
    unsigned char *dst;
    unsigned char *src;
    unsigned int  len;
{
    unsigned char *tmp;
    
    if ((tmp = malloc(len + 1)) == (unsigned char *)NULL){
	return -1;
    }
    strcpy(tmp, src);
    strcpy(dst, tmp);
    free(tmp);
    
    return 0;
}

static int StrnCpy(dst, src, size, len)
    unsigned char *dst;
    unsigned char *src;
    int size;
    unsigned int len;
{
    unsigned char *tmp;
    
    if ((tmp = malloc(len + 1)) == (unsigned char *)NULL){
	return -1;
    }
    strncpy(tmp, src, size);
    strncpy(dst, tmp, size);
    *(dst + size) = '\0';
    free(tmp);
    
    return 0;
}

/*
 * int TSetCursor(col, row)
 *
 * function : move cursor
 *
 * argument : col display column(0 origin)
 *            row display row(0 origin)
 *
 * return  :  0 normal
 *           -1 abnormal
 */
int TSetCursor(col, row)
    int	col;
    int	row;
{
    unsigned char *p;
    
    if (strcmp(p = tgoto(CM, col, row), "OOPS") == 0){
	return -1;
    }
#if 0 /* must I use tputs ? */
    tputs(p, LINES, outchar);
#else
    write(FD_STDOUT, p, strlen(p));
#endif
    return 0;
}

static int TSetCursorItem(col, row, cp)
    int	col;
    int	row;
    int	cp;
{
    TSetCursor(col + cp, row);
    return 0;
}

static int TGetChar(ch)
    unsigned char *ch;
{
    int	i;
    int T_DC_sw = TRUE;
    int T_DC2_sw = TRUE;
    int T_BS_sw = TRUE;
    int T_BS2_sw = TRUE;
    int T_TAB_sw = TRUE;
    int T_BTAB_sw = TRUE;
    int T_ENT_sw = TRUE;
    int	T_FC_sw = TRUE;
    int	T_FC2_sw = TRUE;
    int	T_BC_sw = TRUE;
    int	T_BC2_sw = TRUE;
    int	T_IC_sw = TRUE;
    int	T_IC2_sw = TRUE;
    int	T_PI_sw = TRUE;
    int	T_PI2_sw = TRUE;
    int	T_NI_sw = TRUE;
    int	T_NI2_sw = TRUE;
    int	T_PU_sw = TRUE;
    int	T_PU2_sw = TRUE;
    int	T_PD_sw = TRUE;
    int	T_PD2_sw = TRUE;
    int	T_F01_sw = TRUE;
    int	T_F02_sw = TRUE;
    int	T_F03_sw = TRUE;
    int	T_F04_sw = TRUE;
    int	T_F05_sw = TRUE;
    int	T_F06_sw = TRUE;
    int	T_F07_sw = TRUE;
    int	T_F08_sw = TRUE;
    int	T_F09_sw = TRUE;
    int	T_F10_sw = TRUE;
    int	T_F11_sw = TRUE;
    int	T_F12_sw = TRUE;
    
    read(FD_STDIN, ch, 1);
    
    if (*ch < ' ' || *ch == DEL){
	for (i = 0; TRUE; i++){
	    if (T_DC_sw){
		if (*(T_DC_p + i) == *ch){
		    if (*(T_DC_p + i + 1) == '\0'){
			return T_DC;
		    }
		}else{
		    T_DC_sw = FALSE;
		}
	    }
	    if (T_DC2_sw){
		if (*(T_DC2_p + i) == *ch){
		    if (*(T_DC2_p + i + 1) == '\0'){
			return T_DC;
		    }
		}else{
		    T_DC2_sw = FALSE;
		}
	    }
	    if (T_BS_sw){
		if (*(T_BS_p + i) == *ch){
		    if (*(T_BS_p + i + 1) == '\0'){
			return T_BS;
		    }
		}else{
		    T_BS_sw = FALSE;
		}
	    }
	    if (T_BS2_sw){
		if (*(T_BS2_p + i) == *ch){
		    if (*(T_BS2_p + i + 1) == '\0'){
			return T_BS;
		    }
		}else{
		    T_BS2_sw = FALSE;
		}
	    }
	    if (T_TAB_sw){
		if (*(T_TAB_p + i) == *ch){
		    if (*(T_TAB_p + i + 1) == '\0'){
			return T_TAB;
		    }
		}else{
		    T_TAB_sw = FALSE;
		}
	    }
	    if (T_BTAB_sw){
		if (*(T_BTAB_p + i) == *ch){
		    if (*(T_BTAB_p + i + 1) == '\0'){
			return T_BTAB;
		    }
		}else{
		    T_BTAB_sw = FALSE;
		}
	    }
	    if (T_ENT_sw){
		if (*(T_ENT_p + i) == *ch){
		    if (*(T_ENT_p + i + 1) == '\0'){
			return T_ENT;
		    }
		}else{
		    T_ENT_sw = FALSE;
		}
	    }
	    if (T_FC_sw){
		if (*(T_FC_p + i) == *ch){
		    if (*(T_FC_p + i + 1) == '\0'){
			return T_FC;
		    }
		}else{
		    T_FC_sw = FALSE;
		}
	    }
	    if (T_FC2_sw){
		if (*(T_FC2_p + i) == *ch){
		    if (*(T_FC2_p + i + 1) == '\0'){
			return T_FC;
		    }
		}else{
		    T_FC2_sw = FALSE;
		}
	    }
	    if (T_BC_sw){
		if (*(T_BC_p + i) == *ch){
		    if (*(T_BC_p + i + 1) == '\0'){
			return T_BC;
		    }
		}else{
		    T_BC_sw = FALSE;
		}
	    }
	    if (T_BC2_sw){
		if (*(T_BC2_p + i) == *ch){
		    if (*(T_BC2_p + i + 1) == '\0'){
			return T_BC;
		    }
		}else{
		    T_BC2_sw = FALSE;
		}
	    }
	    if (T_IC_sw){
		if (*(T_IC_p + i) == *ch){
		    if (*(T_IC_p + i + 1) == '\0'){
			return T_IC;
		    }
		}else{
		    T_IC_sw = FALSE;
		}
	    }
	    if (T_IC2_sw){
		if (*(T_IC2_p + i) == *ch){
		    if (*(T_IC2_p + i + 1) == '\0'){
			return T_IC;
		    }
		}else{
		    T_IC2_sw = FALSE;
		}
	    }
	    if (T_PI_sw){
		if (*(T_PI_p + i) == *ch){
		    if (*(T_PI_p + i + 1) == '\0'){
			return T_PI;
		    }
		}else{
		    T_PI_sw = FALSE;
		}
	    }
	    if (T_PI2_sw){
		if (*(T_PI2_p + i) == *ch){
		    if (*(T_PI2_p + i + 1) == '\0'){
			return T_PI;
		    }
		}else{
		    T_PI2_sw = FALSE;
		}
	    }
	    if (T_NI_sw){
		if (*(T_NI_p + i) == *ch){
		    if (*(T_NI_p + i + 1) == '\0'){
			return T_NI;
		    }
		}else{
		    T_NI_sw = FALSE;
		}
	    }
	    if (T_NI2_sw){
		if (*(T_NI2_p + i) == *ch){
		    if (*(T_NI2_p + i + 1) == '\0'){
			return T_NI;
		    }
		}else{
		    T_NI2_sw = FALSE;
		}
	    }
	    if (T_PU_sw){
		if (*(T_PU_p + i) == *ch){
		    if (*(T_PU_p + i + 1) == '\0'){
			return T_PU;
		    }
		}else{
		    T_PU_sw = FALSE;
		}
	    }
	    if (T_PU2_sw){
		if (*(T_PU2_p + i) == *ch){
		    if (*(T_PU2_p + i + 1) == '\0'){
			return T_PU;
		    }
		}else{
		    T_PU2_sw = FALSE;
		}
	    }
	    if (T_PD_sw){
		if (*(T_PD_p + i) == *ch){
		    if (*(T_PD_p + i + 1) == '\0'){
			return T_PD;
		    }
		}else{
		    T_PD_sw = FALSE;
		}
	    }
	    if (T_PD2_sw){
		if (*(T_PD2_p + i) == *ch){
		    if (*(T_PD2_p + i + 1) == '\0'){
			return T_PD;
		    }
		}else{
		    T_PD2_sw = FALSE;
		}
	    }
	    if (T_F01_sw){
		if (*(T_F01_p + i) == *ch){
		    if (*(T_F01_p + i + 1) == '\0'){
			return T_F01;
		    }
		}else{
		    T_F01_sw = FALSE;
		}
	    }
	    if (T_F02_sw){
		if (*(T_F02_p + i) == *ch){
		    if (*(T_F02_p + i + 1) == '\0'){
			return T_F02;
		    }
		}else{
		    T_F02_sw = FALSE;
		}
	    }
	    if (T_F03_sw){
		if (*(T_F03_p + i) == *ch){
		    if (*(T_F03_p + i + 1) == '\0'){
			return T_F03;
		    }
		}else{
		    T_F03_sw = FALSE;
		}
	    }
	    if (T_F04_sw){
		if (*(T_F04_p + i) == *ch){
		    if (*(T_F04_p + i + 1) == '\0'){
			return T_F04;
		    }
		}else{
		    T_F04_sw = FALSE;
		}
	    }
	    if (T_F05_sw){
		if (*(T_F05_p + i) == *ch){
		    if (*(T_F05_p + i + 1) == '\0'){
			return T_F05;
		    }
		}else{
		    T_F05_sw = FALSE;
		}
	    }
	    if (T_F06_sw){
		if (*(T_F06_p + i) == *ch){
		    if (*(T_F06_p + i + 1) == '\0'){
			return T_F06;
		    }
		}else{
		    T_F06_sw = FALSE;
		}
	    }
	    if (T_F07_sw){
		if (*(T_F07_p + i) == *ch){
		    if (*(T_F07_p + i + 1) == '\0'){
			return T_F07;
		    }
		}else{
		    T_F07_sw = FALSE;
		}
	    }
	    if (T_F08_sw){
		if (*(T_F08_p + i) == *ch){
		    if (*(T_F08_p + i + 1) == '\0'){
			return T_F08;
		    }
		}else{
		    T_F08_sw = FALSE;
		}
	    }
	    if (T_F09_sw){
		if (*(T_F09_p + i) == *ch){
		    if (*(T_F09_p + i + 1) == '\0'){
			return T_F09;
		    }
		}else{
		    T_F09_sw = FALSE;
		}
	    }
	    if (T_F10_sw){
		if (*(T_F10_p + i) == *ch){
		    if (*(T_F10_p + i + 1) == '\0'){
			return T_F10;
		    }
		}else{
		    T_F10_sw = FALSE;
		}
	    }
	    if (T_F11_sw){
		if (*(T_F11_p + i) == *ch){
		    if (*(T_F11_p + i + 1) == '\0'){
			return T_F11;
		    }
		}else{
		    T_F11_sw = FALSE;
		}
	    }
	    if (T_F12_sw){
		if (*(T_F12_p + i) == *ch){
		    if (*(T_F12_p + i + 1) == '\0'){
			return T_F12;
		    }
		}else{
		    T_F12_sw = FALSE;
		}
	    }
	    if (T_DC_sw == FALSE &&
	        T_DC2_sw == FALSE &&
		T_BS_sw == FALSE &&
		T_BS2_sw == FALSE &&
		T_TAB_sw == FALSE &&
		T_BTAB_sw == FALSE &&
		T_ENT_sw == FALSE &&
		T_FC_sw == FALSE && 
		T_FC2_sw == FALSE && 
		T_BC_sw == FALSE &&   
		T_BC2_sw == FALSE &&   
		T_IC_sw == FALSE &&   
		T_IC2_sw == FALSE &&   
		T_PI_sw == FALSE &&   
		T_PI2_sw == FALSE &&   
		T_NI_sw == FALSE &&   
		T_NI2_sw == FALSE &&   
		T_PU_sw == FALSE &&   
		T_PU2_sw == FALSE &&   
		T_PD_sw == FALSE &&   
		T_PD2_sw == FALSE &&   
		T_F01_sw == FALSE &&   
		T_F02_sw == FALSE &&   
		T_F03_sw == FALSE &&   
		T_F04_sw == FALSE &&   
		T_F05_sw == FALSE &&   
		T_F06_sw == FALSE &&   
		T_F07_sw == FALSE &&   
		T_F08_sw == FALSE &&   
		T_F09_sw == FALSE &&   
		T_F10_sw == FALSE &&   
		T_F11_sw == FALSE &&   
		T_F12_sw == FALSE){
		return T_IGNORE;
	    }
	    read(FD_STDIN, ch, 1);
	}
    }else if (IsWChar(ch)){
	read(FD_STDIN, ch + 1, 1);
	return T_WCHAR;
    }else if (IsChar(ch)){
	return T_CHAR;
    }else{
	return T_IGNORE;
    }
}

static int IsWChar(ch)
    unsigned char *ch;
{
    switch(CharSet){
      case ASCii:
	return FALSE;
      case JIS:
	return FALSE;
	break;
      case EUC:
	switch (*ch){
	  case EUC_SS2: /* kana is not supported */
	    read(FD_STDIN, ch + 1, 1);
	    return FALSE;
	    break;
	  case EUC_SS3: /* gaiji is not supported */
	    read(FD_STDIN, ch + 1, 1);
	    read(FD_STDIN, ch + 2, 1);
	    return FALSE;
	    break;
	  default:
	    return ((*ch & 0x80) ? TRUE : FALSE);
	    break;
	}
      case SJIS:
	return ((0x81 <= *ch && *ch <= 0x9f) ||
		(0xe0 <= *ch && *ch <= 0xfc) ? TRUE : FALSE);
	break;
      default:
	return FALSE;
	break;
    }
}

static int IsWCharBuf(buf, cp)
    unsigned char *buf;
    int cp;
{
    int i;

    if (cp < 0 || strlen(buf) <= cp){
	return FALSE;
    }
    for (i = 0; i <= cp; i++){
	if (IsWChar(buf + i)){
	    if (i == cp){
		return TRUE;
	    }else{
		i++;
	    }
	}
    }
    return FALSE;
}

static int IsChar(ch)
    unsigned char *ch;
{
    switch (CharSet){
      case ASCii:
	return ((isgraph(*ch) != 0 || *ch == ' ') ? TRUE : FALSE);
	break;
      case JIS:
	return ((isgraph(*ch) != 0 || *ch == ' ') ? TRUE : FALSE);
	break;
      case EUC:
	return ((isgraph(*ch) != 0 || *ch == ' ') ? TRUE : FALSE);
	break;
      case SJIS:
	return ((isgraph(*ch) != 0 || *ch == ' ' ||
		 (0xa0 <= *ch && *ch <= 0xdf)) ? TRUE : FALSE);
	break;
      default:
	return FALSE;
	break;
    }
}

static int IsEnd(status)
    int	status;
{
    if (status == T_PI){
	return TRUE;
    }
    if (status == T_NI){
	return TRUE;
    }
    if (status == T_PU){
	return TRUE;
    }
    if (status == T_PD){
	return TRUE;
    }
    if (status == T_ESC){
	return TRUE;
    }
    if (status == T_TAB){
	return TRUE;
    }
    if (status == T_BTAB){
	return TRUE;
    }
    if (status == T_ENT){
	return TRUE;
    }
    if (status == T_F01){
	return TRUE;
    }
    if (status == T_F02){
	return TRUE;
    }
    if (status == T_F03){
	return TRUE;
    }
    if (status == T_F04){
	return TRUE;
    }
    if (status == T_F05){
	return TRUE;
    }
    if (status == T_F06){
	return TRUE;
    }
    if (status == T_F07){
	return TRUE;
    }
    if (status == T_F08){
	return TRUE;
    }
    if (status == T_F09){
	return TRUE;
    }
    if (status == T_F10){
	return TRUE;
    }
    if (status == T_F11){
	return TRUE;
    }
    if (status == T_F12){
	return TRUE;
    }
    return FALSE;
}

static int ScratchTail(str)
    unsigned char *str;
{
    int i;

    for (i = strlen(str) - 1; 0 <= i && *(str + i) == ' '; i--){
	*(str + i) = '\0';
    }
    return 0;
}

/*
 * int TClear()
 *
 * function : clear screen
 *
 * return :  0 normal
 *          -1 abnormal
 */
int TClear()
{
    write(FD_STDOUT, CL, strlen(CL));
    return 0;
}

#if 0
static int outchar(ch)
    int	ch;
{
    return write(FD_STDOUT, &ch, 1);
}
#endif

/*
 *  screen control functions
 */

/*
 * TSCREEN *TOpenScreen(scrn_name)
 *
 * function : open screen file
 *
 * argument : scrn_name screen name
 *
 * return : TSCREEN pointer normal
 *          NULL            abnormal
 *
 * remarks : search sequence of screen file
 *           1. ${CWD}/scrn_name.scrn
 *           2. ${TSCRNDIR}/scrn_name.scrn
 *           3. /usr/local/lib/libtext/scrn_name.scrn
 */
TSCREEN *TOpenScreen(scrn_name)
    char *scrn_name;
{
    char screen_obj[T_MAXPATHLEN];
    FILE *fp;
    TSCREEN *screen;
    char bufrec[T_OFFSET_STRING+T_MAXSTRLEN+2], *p;
    TITEM *item, *old_item;
    int itemid = 0;

    if (scrn_name == (char *)NULL){
	return (TSCREEN *)NULL;
    }

    fp = (FILE *)NULL;

    if ((strlen(scrn_name) + strlen(".scrn") - 1) > sizeof screen_obj){
	return (TSCREEN *)NULL;
    }
    sprintf(screen_obj, "%s.scrn", scrn_name);
    fp = fopen(screen_obj, "r");
    if (fp == (FILE *)NULL && (p = getenv(TSCRNDIRENV)) != (char *)NULL){
        if ((strlen(p) + strlen("/screen/") + strlen(scrn_name) + 
             strlen(".scrn") - 1) > sizeof screen_obj){
	    return (TSCREEN *)NULL;
        }
	sprintf(screen_obj, "%s/screen/%s.scrn", p, scrn_name);
	fp = fopen(screen_obj, "r");
    }
    if (fp == (FILE *)NULL){
        if ((strlen(TSCRNDEFDIR) + strlen("/screen/") + strlen(scrn_name) + 
             strlen(".scrn") - 1) > sizeof screen_obj){
	    return (TSCREEN *)NULL;
        }
	sprintf(screen_obj, "%s/screen/%s.scrn", TSCRNDEFDIR, scrn_name);
	fp = fopen(screen_obj, "r");
    }
    if (fp == (FILE *)NULL){
	return (TSCREEN *)NULL;
    }

    if ((screen = (TSCREEN *)malloc(sizeof (TSCREEN))) == (TSCREEN *)NULL){
	fclose(fp);
	return (TSCREEN *)NULL;
    }
    if (fgets(bufrec, sizeof bufrec, fp) == (char *)NULL){
	fclose(fp);
	return (TSCREEN *)NULL;
    }
    if (bufrec[T_OFFSET_MODE] != T_MODE_GEOMETRY){
	fclose(fp);
	return (TSCREEN *)NULL;
    }
    bufrec[T_OFFSET_COL+T_SCRN_NUMWIDTH] = '\0';
    bufrec[T_OFFSET_ROW+T_SCRN_NUMWIDTH] = '\0';
    bufrec[T_OFFSET_WIDTH+T_SCRN_NUMWIDTH] = '\0';
    bufrec[T_OFFSET_HEIGHT+T_SCRN_NUMWIDTH] = '\0';
    screen->col = atoi(bufrec+T_OFFSET_COL);
    screen->row = atoi(bufrec+T_OFFSET_ROW);
    screen->width = atoi(bufrec+T_OFFSET_WIDTH);
    screen->height = atoi(bufrec+T_OFFSET_HEIGHT);
    screen->clear = bufrec[T_OFFSET_CLEAR];
    screen->current_itemid = T_ITEMID_TOP;
    screen->top.mode = T_MODE_CONSTANT;
    screen->top.col = 0;
    screen->top.row = 0;
    screen->top.len = 0;
    screen->top.itemid = T_ITEMID_TOP;
    screen->top.attr = T_NORMAL;
    screen->top.string = (char *)NULL;
    screen->top.prev_link = (TITEM *)NULL;
    screen->top.back_link = (TITEM *)NULL;
    screen->top.next_link = (TITEM *)NULL;
    screen->top.forw_link = (TITEM *)NULL;

    old_item = &(screen->top);

    while (fgets(bufrec, sizeof bufrec, fp) != (char *)NULL){
	bufrec[T_OFFSET_COL+T_SCRN_NUMWIDTH] = '\0';
	bufrec[T_OFFSET_ROW+T_SCRN_NUMWIDTH] = '\0';
	bufrec[T_OFFSET_LEN+T_SCRN_NUMWIDTH] = '\0';
	
	if ((item = (TITEM *)malloc(sizeof (TITEM))) == (TITEM *)NULL){
	    fclose(fp);
	    return (TSCREEN *)NULL;
	}

	item->mode = bufrec[T_OFFSET_MODE];
	item->col = atoi(bufrec+T_OFFSET_COL);
	item->row = atoi(bufrec+T_OFFSET_ROW);
	item->len = atoi(bufrec+T_OFFSET_LEN);
	item->itemid = ++itemid;
	item->attr = T_NORMAL;
	if ((item->string = (char *)malloc(item->len + 1)) == (char *)NULL){
	    fclose(fp);
	    return (TSCREEN *)NULL;
	}
#if 0
	if (bufrec[T_OFFSET_MODE] == T_MODE_CONSTANT){
	    bufrec[T_OFFSET_STRING+item->len] = '\0';
	    strncpy(item->string, &bufrec[T_OFFSET_STRING], item->len);
	    *(item->string + item->len) = '\0';
	}else{
	    if (bufrec[strlen(bufrec) - 1] == '\n'){
		bufrec[strlen(bufrec) - 1] = '\0';
	    }
	    strncpy(item->string, &bufrec[T_OFFSET_STRING], item->len);
	    *(item->string + item->len) = '\0';
	}
#else
	strncpy(item->string, &bufrec[T_OFFSET_STRING], item->len);
	*(item->string + item->len) = '\0';
	if (*(item->string + strlen(item->string) - 1) == '\n'){
	    *(item->string + strlen(item->string) - 1) = '\0';
	}
#endif
	item->prev_link = old_item;
	item->back_link = old_item;
	item->next_link = (TITEM *)NULL;
	item->forw_link = (TITEM *)NULL;
	if (screen->current_itemid == T_ITEMID_TOP &&
	    (item->mode == T_MODE_INPUT || item->mode == T_MODE_SELECT)){
	    screen->current_itemid = item->itemid;
	}
	old_item->next_link = item;
	old_item->forw_link = item;
	old_item = item;
    }

    screen->top.prev_link = item;
    screen->top.back_link = item;
    item->next_link = &(screen->top);
    item->forw_link = &(screen->top);

    fclose(fp);

    return (TSCREEN *)screen;
}

/*
 * int TSetItem(screen, itemid, string)
 *
 * function : set value in item
 *
 * argument : screen TSCREEN pointer
 *            itemid item id(which is macro defined as (scrn_name)_itemname in 
 *                   scrn_name.h)
 *            string string to set 
 *
 * return :  0 normal
 *          -1 abnormal
 */
int TSetItem(screen, itemid, string)
    TSCREEN *screen;
    int     itemid;
    char    *string;
{
    TITEM *titemptr;
    int   i;

    if (screen == (TSCREEN *)NULL){
	return -1;
    }
    if (string == (char *)NULL){
	return -1;
    }

    titemptr = &(screen->top);
    for (;;){
	if (titemptr->itemid == itemid){
	    strncpy(titemptr->string, string, titemptr->len);
	    *(titemptr->string + titemptr->len) = '\0';
	    if (titemptr->len < strlen(string)){
		i = 0;
		while (i < titemptr->len){
		    if (IsWChar(titemptr->string + i)){
			if (i == (titemptr->len - 1)){
			    *(titemptr->string + i) = '\0';
			}
			i++;
		    }
		    i++;
		}
	    }
	    ScratchTail(titemptr->string);
	    return 0;
	}
	titemptr = titemptr->forw_link;
	if (titemptr == &(screen->top)){
	    return -1;
	}
    }
}

/*
 * int TOutputScreen(screen)
 *
 * function : output screen
 *
 * argument : screen TSCREEN pointer
 *
 * return :  0 normal
 *          -1 abnormal
 */
int TOutputScreen(screen)
    TSCREEN *screen;
{
    TITEM *titemptr;

    if (screen == (TSCREEN *)NULL){
	return -1;
    }

    switch (screen->clear){
      case T_CLEAR_ALL:
	TClear();
	break;
      case T_CLEAR_PART:
	scrnbuf[0] = '\0';
	TOutputClearScreen(screen, scrnbuf);
	write(FD_STDOUT, scrnbuf, strlen(scrnbuf));
	break;
      case T_CLEAR_NONE:
	break;
      default:
	break;
    }

    scrnbuf[0] = '\0';
    titemptr = (screen->top).forw_link;
    while (titemptr != &(screen->top)){
	TOutputItem(screen, titemptr, scrnbuf);
	titemptr = titemptr->forw_link;
    }
    write(FD_STDOUT, scrnbuf, strlen(scrnbuf));

    return 0;
}

/*
 * int TInputScreen(screen)
 *
 * function : input screen
 *
 * argument : screen TSCREEN pointer
 *
 * return : action to terminate input
 *
 *        T_PU     PageUp
 *        T_PD     PageDown
 *        T_ENT    ENTER
 *        T_F01    FUNCTION01
 *        T_F02
 *        T_F03
 *        T_F04
 *        T_F05
 *        T_F06
 *        T_F07
 *        T_F08
 *        T_F09
 *        T_F10
 *        T_F11
 *        T_F12    FUNCTION12
 *        
 *        -1 abnormal
 */
int TInputScreen(screen)
    TSCREEN *screen;
{
    TITEM *titemptr, *titemptr_old;
    int   ret;

    if (screen == (TSCREEN *)NULL){
	return -1;
    }
    
    titemptr = TGetTitemPointer(screen, screen->current_itemid);
    if (titemptr->mode == T_MODE_SELECT){
	scrnbuf[0] = '\0';
	TOutputItem(screen, titemptr, scrnbuf);
	write(FD_STDOUT, scrnbuf, strlen(scrnbuf));
	ret = TInputString(screen->col+titemptr->col, screen->row+titemptr->row, 0, titemptr->string);
    }else{
	ret = TInputString(screen->col+titemptr->col, screen->row+titemptr->row, titemptr->len, titemptr->string);
    }
    while (ret != T_PU &&
	   ret != T_PD &&
	   ret != T_ENT &&
	   ret != T_F01 &&
	   ret != T_F02 &&
	   ret != T_F03 &&
	   ret != T_F04 &&
	   ret != T_F05 &&
	   ret != T_F06 &&
	   ret != T_F07 &&
	   ret != T_F08 &&
	   ret != T_F09 &&
	   ret != T_F10 &&
	   ret != T_F11 &&
	   ret != T_F12){

	titemptr_old = titemptr;
	switch (ret){
	  case T_FC:
	    titemptr = TGetForwTitemPointer(titemptr);
	    screen->current_itemid = titemptr->itemid;
	    break;
	  case T_BC:
	    titemptr = TGetBackTitemPointer(titemptr);
	    screen->current_itemid = titemptr->itemid;
	    break;
	  case T_DC:
	    break;
	  case T_IC:
	    break;
	  case T_BS:
	    break;
	  case T_CHAR:
	    titemptr = TGetForwTitemPointer(titemptr);
	    screen->current_itemid = titemptr->itemid;
	    break;
	  case T_WCHAR:
	    titemptr = TGetForwTitemPointer(titemptr);
	    screen->current_itemid = titemptr->itemid;
	    break;
	  case T_IGNORE:
	    break;
	  case T_PI:
	    titemptr = TGetPrevTitemPointer(titemptr);
	    screen->current_itemid = titemptr->itemid;
	    break;
	  case T_NI:
	    titemptr = TGetNextTitemPointer(titemptr);
	    screen->current_itemid = titemptr->itemid;
	    break;
	  case T_ESC:
	    break;
	  case T_TAB:
	    titemptr = TGetForwTitemPointer(titemptr);
	    screen->current_itemid = titemptr->itemid;
	    break;
	  case T_BTAB:
	    titemptr = TGetBackTitemPointer(titemptr);
	    screen->current_itemid = titemptr->itemid;
	    break;
	  default:
	    break;
	}
	if (titemptr_old->mode == T_MODE_SELECT){
	    scrnbuf[0] = '\0';
	    TOutputItem(screen, titemptr_old, scrnbuf);
	    write(FD_STDOUT, scrnbuf, strlen(scrnbuf));
	}
	if (titemptr->mode == T_MODE_SELECT){
	    scrnbuf[0] = '\0';
	    TOutputItem(screen, titemptr, scrnbuf);
	    write(FD_STDOUT, scrnbuf, strlen(scrnbuf));
	    ret = TInputString(screen->col+titemptr->col, screen->row+titemptr->row, 0, titemptr->string);
	}else{
	    ret = TInputString(screen->col+titemptr->col, screen->row+titemptr->row, titemptr->len, titemptr->string);
	}
    }
    return ret;
}

/*
 * int TIOScreen(screen)
 *
 * function : input/output screen
 *
 * argument : screen TSCREEN pointer
 *
 * return : action to terminate input
 *
 *        T_PU     PageUp
 *        T_PD     PageDown
 *        T_ENT    ENTER
 *        T_F01    FUNCTION01
 *        T_F02
 *        T_F03
 *        T_F04
 *        T_F05
 *        T_F06
 *        T_F07
 *        T_F08
 *        T_F09
 *        T_F10
 *        T_F11
 *        T_F12    FUNCTION12
 *        
 *        -1 abnormal
 */
int TIOScreen(screen)
    TSCREEN *screen;
{
    if (screen == (TSCREEN *)NULL){
	return -1;
    }
    
    TOutputScreen(screen);
    return TInputScreen(screen);
}

static TITEM *TGetTitemPointer(screen, itemid)
    TSCREEN *screen;
    int itemid;
{
    TITEM *titemptr;

    titemptr = &(screen->top);
    for (;;){
	if (titemptr->itemid == itemid){
	    return titemptr;
	}
	titemptr = titemptr->forw_link;
	if (titemptr == &(screen->top)){
	    return (TITEM *)NULL;
	}
    }
}

static TITEM *TGetForwTitemPointer(titem)
    TITEM *titem;
{
    TITEM *titemptr;

    titemptr = titem->forw_link;
    while ((titemptr->mode == T_MODE_CONSTANT || titemptr->mode == T_MODE_OUTPUT) &&
	   titemptr != titem){
	titemptr = titemptr->forw_link;
    }
    return (TITEM *)titemptr;
}

static TITEM *TGetNextTitemPointer(titem)
    TITEM *titem;
{
    TITEM *titemptr;
    int i = 0;

    titemptr = titem->next_link;
    while ((titemptr->mode == T_MODE_CONSTANT || titemptr->mode == T_MODE_OUTPUT) &&
	   ++i < T_MAXITEMNUM){
	titemptr = titemptr->next_link;
    }
    if (T_MAXITEMNUM <= i){
	return (TITEM *)titem;
    }else{
	return (TITEM *)titemptr;
    }
}

static TITEM *TGetBackTitemPointer(titem)
    TITEM *titem;
{
    TITEM *titemptr;

    titemptr = titem->back_link;
    while ((titemptr->mode == T_MODE_CONSTANT || titemptr->mode == T_MODE_OUTPUT) &&
	   titemptr != titem){
	titemptr = titemptr->back_link;
    }
    return (TITEM *)titemptr;
}

static TITEM *TGetPrevTitemPointer(titem)
    TITEM *titem;
{
    TITEM *titemptr;
    int i = 0;

    titemptr = titem->prev_link;
    while ((titemptr->mode == T_MODE_CONSTANT || titemptr->mode == T_MODE_OUTPUT)&&
	   ++i < T_MAXITEMNUM){
	titemptr = titemptr->prev_link;
    }
    if (T_MAXITEMNUM <= i){
	return (TITEM *)titem;
    }else{
	return (TITEM *)titemptr;
    }
}

/*
 * char *TGetItem(screen, itemid)
 *
 * function : get item value
 *
 * argument : screen TSCREEN pointer
 *            itemid item id
 *
 * return : pointer to buffer stored item value normal
 *          NULL                                abnormal
 */
char *TGetItem(screen, itemid)
    TSCREEN *screen;
    int     itemid;
{
    TITEM *titemptr;

    if (screen == (TSCREEN *)NULL){
	return (char *)NULL;
    }
    
    titemptr = &(screen->top);
    for (;;){
	if (titemptr->itemid == itemid){
	    return titemptr->string;
	}
	titemptr = titemptr->forw_link;
	if (titemptr == &(screen->top)){
	    return (char *)NULL;
	}
    }
}

/*
 * int TCloseScreen(screen)
 *
 * function : close screen file
 *
 * argument : screen TSCREEN pointer
 *
 * return :  0 normal
 *          -1 abnormal
 */
int TCloseScreen(screen)
    TSCREEN *screen;
{
    TITEM *titemptr, *save;

    if (screen == (TSCREEN *)NULL){
	return -1;
    }
    
    titemptr = (screen->top).forw_link;
    for (;;){
	free(titemptr->string);
	save = titemptr->forw_link;
	free(titemptr);
	titemptr = save;
	if (titemptr == &(screen->top)){
	    break;
	}
    }
    free(screen);
    return 0;
}

/*
 * int TSetCurrentCursor(screen, itemid)
 *
 * function : move cursor on item
 *
 * argument : screen TSCREEN pointer
 *            itemid item id
 *
 * return :  0 normal
 *          -1 abnormal
 */
int TSetCurrentCursor(screen, itemid)
    TSCREEN *screen;
    int     itemid;
{
    TITEM *titemptr;

    if (screen == (TSCREEN *)NULL){
	return -1;
    }
    
    titemptr = TGetTitemPointer(screen, itemid);
    if (titemptr->mode == T_MODE_INPUT || titemptr->mode == T_MODE_SELECT){
	screen->current_itemid = itemid;
	return 0;
    }
    return -1;
}

/*
 * int TSetItemMode(screen, itemid, mode)
 *
 * function : set screen mode
 *
 * argument : screen TSCREEN pointer
 *            itemid item id
 *            mode   screen mode
 *                   T_MODE_INPUT    input mode
 *                   T_MODE_OUTPUT   output mode
 *                   T_MODE_CONSTANT constant string mode
 *                   T_MODE_SELECT   select mode
 *
 * return :  0 normal
 *          -1 abnormal
 */
int TSetItemMode(screen, itemid, mode)
    TSCREEN *screen;
    int     itemid;
    int     mode;
{
    if (screen == (TSCREEN *)NULL){
	return -1;
    }

    TGetTitemPointer(screen, itemid)->mode = mode;
    return 0;
}

/*
 * int TSetNextItem(screen, itemid, next_itemid)
 *
 * function : set next item by down cursor
 *
 * argument : screen      TSCREEN pointer
 *            itemid      item id
 *            next_itemid item id to move cursor
 *
 * return :  0 normal
 *          -1 abnormal
 */
int TSetNextItem(screen, itemid, next_itemid)
    TSCREEN *screen;
    int     itemid;
    int     next_itemid;
{
    if (screen == (TSCREEN *)NULL){
	return -1;
    }
    
    TGetTitemPointer(screen, itemid)->next_link = TGetTitemPointer(screen, next_itemid);
    return 0;
}

/*
 * int TSetPrevItem(screen, itemid, next_itemid)
 *
 * function : set next item by up cursor
 *
 * argument : screen      TSCREEN pointer
 *            itemid      item id
 *            next_itemid item id to move cursor
 *
 * return :  0 normal
 *          -1 abnormal
 */
int TSetPrevItem(screen, itemid, prev_itemid)
    TSCREEN *screen;
    int     itemid;
    int     prev_itemid;
{
    if (screen == (TSCREEN *)NULL){
	return -1;
    }
    
    TGetTitemPointer(screen, itemid)->prev_link = TGetTitemPointer(screen, prev_itemid);
    return 0;
}

/*
 * int TSetItemAttr(screen, itemid, attr)
 *
 * function : set item attribute
 *
 * argument : screen TSCREEN pointer
 *            itemid item id
 *            attr   attribute
 *                   T_BOLD      bold
 *                   T_REVERSE   reverse
 *                   T_UNDERLINE underline
 *                   T_NORMAL    normal
 *                   T_BLACK     black
 *                   T_RED       red
 *                   T_GREEN     green
 *                   T_YELLOW    yellow
 *                   T_BLUE      blue
 *                   T_PURPLE    purple
 *                   T_CYAN      ctan
 *                   T_WHITE     white
 *
 * return :  0 normal
 *          -1 abnormal
 */
int TSetItemAttr(screen, itemid, attr)
    TSCREEN *screen;
    int     itemid;
    int     attr;
{
    TITEM *titemptr;

    if (screen == (TSCREEN *)NULL){
	return -1;
    }

    titemptr = TGetTitemPointer(screen, itemid);
    titemptr->attr = 0;
    
    if (attr & T_BLACK){
	titemptr->attr |= T_BLACK;
    }
    if (attr & T_RED){
	titemptr->attr |= T_RED;
    }
    if (attr & T_GREEN){
	titemptr->attr |= T_GREEN;
    }
    if (attr & T_YELLOW){
	titemptr->attr |= T_YELLOW;
    }
    if (attr & T_BLUE){
	titemptr->attr |= T_BLUE;
    }
    if (attr & T_PURPLE){
	titemptr->attr |= T_PURPLE;
    }
    if (attr & T_CYAN){
	titemptr->attr |= T_CYAN;
    }
    if (attr & T_WHITE){
	titemptr->attr |= T_WHITE;
    }
    if (attr & T_BOLD){
	titemptr->attr |= T_BOLD;
    }
    if (attr & T_REVERSE){
	titemptr->attr |= T_REVERSE;
    }
    if (attr & T_UNDERLINE){
	titemptr->attr |= T_UNDERLINE;
    }
    if (attr & T_NORMAL){
	titemptr->attr |= T_NORMAL;
    }
	
    return 0;
}

static int TOutputClearScreen(screen, scrnbuf)
    TSCREEN *screen;
    char    *scrnbuf;
{
    int row, width;
    unsigned char *p;

    for (row = screen->row; row < (screen->row + screen->height); row++){
	if (strcmp(p = tgoto(CM, screen->col, row), "OOPS") == 0){
	    return -1;
	}
	strcat(scrnbuf, p);
	for (width = 0; width < screen->width; width++){
	    strcat(scrnbuf, " ");
	}
    }

    return 0;
}

static int TOutputItem(screen, titemptr, scrnbuf)
    TSCREEN *screen;
    TITEM   *titemptr;
    char    *scrnbuf;
{
    unsigned char *p;

    strcat(scrnbuf, T_NORMAL_p);

    if (strcmp(p = tgoto(CM, screen->col+titemptr->col, screen->row+titemptr->row), "OOPS") == 0){
	return -1;
    }
    strcat(scrnbuf, p);

    if (titemptr->attr & T_NORMAL){
	if (titemptr->mode == T_MODE_SELECT && screen->current_itemid == titemptr->itemid){
	    strcat(scrnbuf, T_REVERSE_p);
	}
    }else{
	if (titemptr->attr & T_BLACK){
	    strcat(scrnbuf, T_BLACK_p);
	}else if (titemptr->attr & T_RED){
	    strcat(scrnbuf, T_RED_p);
	}else if (titemptr->attr & T_GREEN){
	    strcat(scrnbuf, T_GREEN_p);
	}else if (titemptr->attr & T_YELLOW){
	    strcat(scrnbuf, T_YELLOW_p);
	}else if (titemptr->attr & T_BLUE){
	    strcat(scrnbuf, T_BLUE_p);
	}else if (titemptr->attr & T_PURPLE){
	    strcat(scrnbuf, T_PURPLE_p);
	}else if (titemptr->attr & T_CYAN){
	    strcat(scrnbuf, T_CYAN_p);
	}else if (titemptr->attr & T_WHITE){
	    strcat(scrnbuf, T_WHITE_p);
	}
	if (titemptr->attr & T_BOLD){
	    strcat(scrnbuf, T_BOLD_p);
	}
	if ((titemptr->attr & T_REVERSE) ||
	    (titemptr->mode == T_MODE_SELECT && screen->current_itemid == titemptr->itemid)){
	    strcat(scrnbuf, T_REVERSE_p);
	}
	if (titemptr->attr & T_UNDERLINE){
	    strcat(scrnbuf, T_UNDERLINE_p);
	}
    }
    strcat(scrnbuf, titemptr->string);
    strcat(scrnbuf, T_NORMAL_p);
    
    return 0;
}

/*
 * int TGetSelectedItem(screen)
 *
 * function : get selected item id
 *
 * argument : screen TSCREEN pointer
 *
 * return : item id normal
            -1      abnormal
 */
int TGetSelectedItem(screen)
    TSCREEN *screen;
{
    if (screen == (TSCREEN *)NULL){
	return -1;
    }
    
    if (TGetTitemPointer(screen, screen->current_itemid)->mode == T_MODE_SELECT){
	return screen->current_itemid;
    }else{
	return -1;
    }
}

/*
 * int TSetScreenPosition(screen, col, row)
 *
 * function : set screen position
 *
 * argument : screen TSCREEN pointer
 *            col    screen column(0 origin)
 *            row    screen row(0 origin)
 *
 * return :  0 normal
 *          -1 abnormal
 */
int TSetScreenPosition(screen, col, row)
    TSCREEN *screen;
    int     col;
    int     row;
{
    if (screen == (TSCREEN *)NULL){
	return -1;
    }
    
    screen->col = col;
    screen->row = row;
    return 0;
}

/*
 * int TSetScreenSize(screen, width, height)
 *
 * function : set screen size
 *
 * argument : screen TSCREEN pointer
 *            width  screen width
 *            height screen height
 *
 * return :  0 normal
 *          -1 abnormal
 */
int TSetScreenSize(screen, width, height)
    TSCREEN *screen;
    int     width;
    int     height;
{
    if (screen == (TSCREEN *)NULL){
	return -1;
    }
    
    screen->width = width;
    screen->height = height;
    return 0;
}

/*
 * int TGetItemPosition(screen, itemid, col, row)
 *
 * function : get item position
 *
 * argument : screen TSCREEN pointer
 *            itemid item id
 *            col    int pointer to store item column
 *            row    int pointer to store item row
 *
 * return :  0 normal
 *          -1 abnormal
 */
int TGetItemPosition(screen, itemid, col, row)
    TSCREEN *screen;
    int     itemid;
    int     *col;
    int     *row;
{
    if (screen == (TSCREEN *)NULL){
	return -1;
    }
    if (col == (int *)NULL){
	return -1;
    }
    if (row == (int *)NULL){
	return -1;
    }
    
    *col = TGetTitemPointer(screen, itemid)->col;
    *row = TGetTitemPointer(screen, itemid)->row;
    return 0;
}

/*
 * int TGetScreenPosition(screen, col, row)
 *
 * function : get screen position
 *
 * argument : screen TSCREEN pointer
 *            col    int pointer to store screen column
 *            row    int pointer to store screen row
 *
 * return :  0 normal
 *          -1 abnormal
 */
int TGetScreenPosition(screen, col, row)
    TSCREEN *screen;
    int     *col;
    int     *row;
{
    if (screen == (TSCREEN *)NULL){
	return -1;
    }
    if (col == (int *)NULL){
	return -1;
    }
    if (row == (int *)NULL){
	return -1;
    }
    
    *col = screen->col;
    *row = screen->row;
    return 0;
}

/*
 * int TGetScreenSize(screen, width, height)
 *
 * function : get screen size
 *
 * argument : screen TSCREEN pointer
 *            width  int pointer to store screen width
 *            height int pointer to store screen height
 *
 * return :  0 normal
 *          -1 abnormal
 */
int TGetScreenSize(screen, width, height)
    TSCREEN *screen;
    int     *width;
    int     *height;
{
    if (screen == (TSCREEN *)NULL){
	return -1;
    }
    if (width == (int *)NULL){
	return -1;
    }
    if (height == (int *)NULL){
	return -1;
    }
    
    *width = screen->width;
    *height = screen->height;
    return 0;
}

/*
 * int TSetClearMode(screen, mode)
 *
 * function : set screen clear mode
 *
 * argument : screen TSCREEN pointer
 *            mode   clear mode
 *                   T_CLEAR_ALL  clear all
 *                   T_CLEAR_PART clear partiality
 *                   T_CLEAR_NONE no clear
 *
 * return :  0 normal
 *          -1 abnormal
 */
int TSetClearMode(screen, mode)
    TSCREEN *screen;
    int     mode;
{
    if (screen == (TSCREEN *)NULL){
	return -1;
    }
    
    screen->clear = mode;
    return 0;
}

/*
 * int TClearScreen(screen)
 *
 * function : clear all screen
 *
 * argument : screen TSCREEN pointer
 *
 * return :  0 normal
 *          -1 abnormal
 */
int TClearScreen(screen)
    TSCREEN *screen;
{
    TITEM *titemptr;

    if (screen == (TSCREEN *)NULL){
	return -1;
    }

    titemptr = &(screen->top);
    for (;;){
	if (titemptr->mode != T_MODE_CONSTANT){
	    *(titemptr->string) = '\0';
	}
	titemptr = titemptr->forw_link;
	if (titemptr == &(screen->top)){
	    return 0;
	}
    }
}
