/*
 *  Text screen control library
 */
#define TSCRNDIRENV "TSCRNDIR"
#define TSCRNDEFDIR "/usr/local/tscrn"

#define T_MAXPATHLEN 256
#define T_MAXSTRLEN  256
#define T_MAXITEMNUM 256
#define T_MAXSCRNBUF 8192

#define T_OFFSET_MODE     0
#define T_OFFSET_COL      2
#define T_OFFSET_ROW      6
#define T_OFFSET_WIDTH    10
#define T_OFFSET_HEIGHT   14
#define T_OFFSET_CLEAR    18
#define T_OFFSET_LEN      10
#define T_OFFSET_STRING   14
#define T_SCRN_NUMWIDTH   3

#define T_BOLD              1 /* md */
#define T_REVERSE           2 /* mr */
#define T_UNDERLINE         4 /* us */
#define T_NORMAL            8 /* me */
#define T_BLACK             0x10
#define T_BLACK_STR         "\033[30m"
#define T_RED               0x20
#define T_RED_STR           "\033[31m"
#define T_GREEN             0x40
#define T_GREEN_STR         "\033[32m"
#define T_YELLOW            0x80
#define T_YELLOW_STR        "\033[33m"
#define T_BLUE              0x100
#define T_BLUE_STR          "\033[34m"
#define T_PURPLE            0x200
#define T_PURPLE_STR        "\033[35m"
#define T_CYAN              0x400
#define T_CYAN_STR          "\033[36m"
#define T_WHITE             0x800
#define T_WHITE_STR         "\033[37m"

#define T_FC      1       /* kr */
#define T_FC_STR "\006"   /* ^F */
#define T_BC      2       /* kl */
#define T_BC_STR "\002"   /* ^B */
#define T_DC      3       /* kD */
#define T_DC_STR "\004"   /* ^D */
#define T_IC      4       /* kI */
#define T_IC_STR "\017"   /* ^O */
#define T_BS      5       /* kb */
#define T_BS_STR "\010"   /* ^H */
#define T_CHAR    6
#define T_WCHAR   7
#define T_IGNORE  8
#define T_PI      9       /* ku */
#define T_PI_STR "\020"   /* ^P */
#define T_NI      10      /* kd */
#define T_NI_STR "\016"   /* ^N */
#define T_PU      11      /* kP */
#define T_PU_STR "\026"   /* ^V */
#define T_PD      12      /* kN */
#define T_PD_STR "\032"   /* ^Z */
#define T_ESC     13
#define T_ESC_STR "\033"  /* ^[ */
#define T_TAB     14
#define T_TAB_STR "\011"  /* ^I */
#define T_BTAB    15
#define T_BTAB_STR "\025" /* ^U */
#define T_ENT     16
#define T_ENT_STR "\012" /* ^M */
#define T_F01     17     /* k1 */
#define T_F02     18     /* k2 */
#define T_F03     19     /* k3 */
#define T_F04     20     /* k4 */
#define T_F05     21     /* k5 */
#define T_F06     22     /* k6 */
#define T_F07     23     /* k7 */
#define T_F08     24     /* k8 */
#define T_F09     25     /* k9 */
#define T_F10     26     /* k0 */
#define T_F11     27     /* F1 */
#define T_F12     28     /* F2 */

#define T_ITEMID_TOP    0

#define T_MODE_GEOMETRY 'G'
#define T_MODE_INPUT    'I'
#define T_MODE_OUTPUT   'O'
#define T_MODE_CONSTANT 'C'
#define T_MODE_SELECT   'S'

#define T_CLEAR_ALL     'Y'
#define T_CLEAR_PART    'P'
#define T_CLEAR_NONE    'N'

typedef struct st_item{
    int    mode;
    int    col;
    int    row;
    int    len;
    int    itemid;
    int    attr;
    char   *string;
    struct st_item *prev_link; /* up cursor you can change */
    struct st_item *back_link; /* left cursor you can't change */
    struct st_item *next_link; /* down cursor you can change */
    struct st_item *forw_link; /* right cursor you can't change */
} TITEM;

typedef struct st_tscreen{
    int col;
    int row;
    int width;
    int height;
    int clear;
    int current_itemid;
    struct st_item top;
} TSCREEN;

extern int TInitTextLib();
extern int TIOString();
extern int TOutputString();
extern int TInputString();
extern int TSetCursor();
extern int TClear();

extern TSCREEN *TOpenScreen();
extern int TSetItem();
extern int TOutputScreen();
extern int TInputScreen();
extern int TIOScreen();
extern char *TGetItem();
extern int TCloseScreen();
extern int TSetCurrentCursor();
extern int TSetItemMode();
extern int TSetNextItem();
extern int TSetPrevItem();
extern int TSetItemAttr();
extern int TGetSelectedItem();
extern int TSetScreenPosition();
extern int TSetScreenSize();
extern int TGetItemPosition();
extern int TGetScreenPosition();
extern int TGetScreenSize();
extern int TSetClearMode();
extern int TClearScreen();
